"""Python module for accessing SkyQ box and EPG, and sending commands."""
import logging
from datetime import datetime

from .classes.app import AppInformation
from .classes.channel import ChannelInformation
from .classes.channelepg import ChannelEPGInformation
from .classes.device import DeviceInformation, TransportInfo
from .classes.deviceaccess import DeviceAccess
from .classes.favourite import FavouriteInformation
from .classes.media import MediaInformation
from .classes.programme import Programme
from .classes.recordings import RecordingsInformation
from .const import (
    ALLRECORDINGS,
    COMMANDS,
    EPG_ERROR_NO_DATA,
    EPG_ERROR_PAST_END,
    SKY_STATE_OFF,
    SKY_STATE_ON,
    SKY_STATE_STANDBY,
    SKY_STATE_UNSUPPORTED,
    TERRITORIES,
    UNSUPPORTED_DEVICES,
    URL_PREFIX,
)

_LOGGER = logging.getLogger(__name__)


class SkyQRemote:
    """SkyQRemote is the instantiation of the SKYQ remote ccontrol."""

    commands = COMMANDS

    def __init__(self, host, epg_cache_len=20, port=49160, json_port=9006):
        """Stand up a new SkyQ box."""
        self._remote_setup = False
        self._device_type = None
        self._host = host
        self._override_country = None
        self._channel = None
        self._programme = None
        self._recorded_programme = None
        self._last_programme_epg = None
        self._last_pvr_id = None
        self._error = False
        self._channellist = None

        self._app_information = None
        self._channel_information = None
        self._device_information = None
        self._channel_epg_information = None
        self._favourite_information = None
        self._media_information = None
        self._recordings_information = None

        self._remote_config = _RemoteConfig(host, port, json_port, epg_cache_len)

        self._setup_remote()

    @property
    def device_setup(self):
        """Get the dev ice setp state."""
        return self._remote_setup

    @property
    def device_type(self):
        """Get the device type of the Sky Q box."""
        return self._device_type

    def power_status(self) -> str:
        """Get the power status of the Sky Q box."""
        if not self._remote_setup:
            self._setup_remote()

        system_info = self._device_information.get_system_information()

        if system_info is None:
            return SKY_STATE_OFF
        if "activeStandby" in system_info and system_info["activeStandby"] is True:
            return SKY_STATE_STANDBY

        return SKY_STATE_ON

    def get_current_state(self):
        """Get current state of the SkyQ box."""
        if not self._remote_setup:
            self._setup_remote()

        if self._device_type in UNSUPPORTED_DEVICES:
            return TransportInfo(SKY_STATE_UNSUPPORTED, None, None)

        return self._device_information.get_transport_information()

    def get_active_application(self):
        """Get the active application on Sky Q box."""
        if not self._app_information:
            self._app_information = AppInformation(self._remote_config)

        return self._app_information.get_active_application()

    def get_current_media(self):
        """Get the currently playing media on the SkyQ box."""
        if not self._media_information:
            self._media_information = MediaInformation(self._remote_config)

        return self._media_information.get_current_media()

    def get_epg_data(self, sid, epg_date, days=2):
        """Get EPG data for the specified channel/date."""
        if not self._channel_epg_information:
            self._channel_epg_information = ChannelEPGInformation(self._remote_config)

        return self._channel_epg_information.get_epg_data(sid, epg_date, days)

    def get_programme_from_epg(self, sid, epg_date, query_date):
        """Get programme from EPG for specfied time and channel."""
        sidint = 0
        try:
            sidint = int(sid)
        except ValueError:
            if not self._error:
                self._error = True
                _LOGGER.info(
                    "I0010 - Programme data not found for host: %s/%s sid: %s : %s",
                    self._host,
                    self._override_country,
                    sid,
                    epg_date,
                )
                return EPG_ERROR_NO_DATA

            self._error = False

        programme_epg = f'{sidint} {epg_date.strftime("%Y%m%d")}'
        if (
            self._last_programme_epg == programme_epg
            and query_date < self._programme.endtime
        ):
            return self._programme

        epg_data = self.get_epg_data(sidint, epg_date)

        if len(epg_data.programmes) == 0:
            if not self._error:
                self._error = True
                _LOGGER.info(
                    "I0020 - Programme data not found for host: %s/%s sid: %s : %s",
                    self._host,
                    self._override_country,
                    sid,
                    epg_date,
                )
            return EPG_ERROR_NO_DATA

        self._error = False

        try:
            programme = next(
                p
                for p in epg_data.programmes
                if p.starttime <= query_date and p.endtime >= query_date
            )

            self._programme = programme
            self._last_programme_epg = programme_epg
            return programme

        except StopIteration:
            return EPG_ERROR_PAST_END

    def get_current_live_tv_programme(self, sid):
        """Get current live programme on the specified channel."""
        try:
            query_date = datetime.utcnow()
            programme = self.get_programme_from_epg(sid, query_date, query_date)
            if not isinstance(programme, Programme):
                return None

            return programme
        except Exception as err:  # pylint: disable=broad-except
            _LOGGER.exception(
                "X0010 - Error occurred: %s : %s : %s", self._host, sid, err
            )
            return None

    def get_recordings(self, status=ALLRECORDINGS, limit=1000, offset=0):
        """Get the list of available Recordings."""
        if not self._recordings_information:
            self._recordings_information = RecordingsInformation(self._remote_config)

        return self._recordings_information.get_recordings(status, limit, offset)

    def get_recording(self, pvrid):
        """Get the recording details."""
        if not self._recordings_information:
            self._recordings_information = RecordingsInformation(self._remote_config)

        if self._last_pvr_id == pvrid:
            return self._recorded_programme
        self._last_pvr_id = pvrid

        self._recorded_programme = self._recordings_information.get_recording(pvrid)
        return self._recorded_programme

    def get_quota(self):
        """Retrieve the available storage quota."""
        if not self._recordings_information:
            self._recordings_information = RecordingsInformation(self._remote_config)

        return self._recordings_information.get_quota()

    def book_recording(self, eid, series=False):
        """Book recording for specified item."""
        if not self._recordings_information:
            self._recordings_information = RecordingsInformation(self._remote_config)

        return self._recordings_information.book_recording(eid, series)

    def book_ppv_recording(self, eid, offerref):
        """Book recording for specified item."""
        if not self._recordings_information:
            self._recordings_information = RecordingsInformation(self._remote_config)

        return self._recordings_information.book_ppv_recording(eid, offerref)

    def series_link(self, pvrid, linkon=True):
        """Book recording for specified item."""
        if not self._recordings_information:
            self._recordings_information = RecordingsInformation(self._remote_config)

        return self._recordings_information.series_link(pvrid, linkon)

    def recording_keep(self, pvrid, keepon=True):
        """Keep the recording."""
        if not self._recordings_information:
            self._recordings_information = RecordingsInformation(self._remote_config)

        return self._recordings_information.recording_keep(pvrid, keepon)

    def recording_lock(self, pvrid, lockon=True):
        """Lock the recording."""
        if not self._recordings_information:
            self._recordings_information = RecordingsInformation(self._remote_config)

        return self._recordings_information.recording_lock(pvrid, lockon)

    def recording_delete(self, pvrid, deleteon=True):
        """Delete the recording."""
        if not self._recordings_information:
            self._recordings_information = RecordingsInformation(self._remote_config)

        return self._recordings_information.recording_delete(pvrid, deleteon)

    def recording_erase(self, pvrid):
        """Delete the recording."""
        if not self._recordings_information:
            self._recordings_information = RecordingsInformation(self._remote_config)

        return self._recordings_information.recording_erase(pvrid)

    def recording_erase_all(self):
        """Delete the reording."""
        if not self._recordings_information:
            self._recordings_information = RecordingsInformation(self._remote_config)

        return self._recordings_information.recording_erase_all()

    def recording_set_last_played_position(self, pvrid, pos):
        """Set the last played position for specified item."""
        if not self._recordings_information:
            self._recordings_information = RecordingsInformation(self._remote_config)

        return self._recordings_information.recording_set_last_played_position(
            pvrid, pos
        )

    def get_device_information(self):
        """Get the device information from the SkyQ box."""
        if not self._device_information:
            self._device_information = DeviceInformation(self._remote_config)

        device_info = self._device_information.get_device_information(
            self._override_country
        )
        if not device_info:
            return None

        self._remote_config.set_device_info(device_info)
        return device_info

    def get_channel_list(self):
        """Get Channel list for Sky Q box."""
        if not self._channel_information:
            self._channel_information = ChannelInformation(self._remote_config)

        self._channellist = self._channel_information.get_channel_list()
        return self._channellist

    def get_channel_info(self, channel_no):
        """Retrieve channel information for specified channelNo."""
        if not self._channel_information:
            self._channel_information = ChannelInformation(self._remote_config)

        return self._channel_information.get_channel_info(channel_no)

    def get_favourite_list(self):
        """Retrieve the list of favourites."""
        if not self._favourite_information:
            self._favourite_information = FavouriteInformation(self._remote_config)

        if not self._channellist:
            self.get_channel_list()
        return self._favourite_information.get_favourite_list(self._channellist)

    def press(self, sequence):
        """Issue the specified sequence of commands to SkyQ box."""
        self._remote_config.device_access.press(sequence)

    def set_overrides(
        self, override_country=None, test_channel=None, json_port=None, port=None
    ):
        """Override various items."""
        if override_country:
            self._override_country = override_country
        if test_channel:
            self._remote_config.test_channel = test_channel
        if json_port:
            self._remote_config.json_port = json_port
        if port:
            self._remote_config.port = port

    def _setup_remote(self):
        device_info = self.get_device_information()
        if not device_info:
            return

        self._remote_setup = True
        self._device_type = self._remote_config.device_info.deviceType


class _RemoteConfig:
    device_access = None
    host = 0
    port = 0
    json_port = 0
    epg_cache_len = 0
    remote_country = ""
    test_channel = 0
    epg_cache_len = 0
    device_info = None

    def __init__(
        self,
        host,
        port,
        json_port,
        epg_cache_len,
        device_access=None,
        remote_country=None,
        test_channel=None,
        device_info=None,
    ):
        self.host = host
        self.port = port
        self.json_port = json_port
        self.device_access = device_access
        self.remote_country = remote_country
        self.territory = None
        self.test_channel = test_channel
        self.epg_cache_len = epg_cache_len
        self.device_access = DeviceAccess(host, json_port, port)
        self.device_info = device_info
        self.territory = None
        self.url_prefix = None

    def set_device_info(self, device_info):
        """Initilise the device info for the Sky Q box."""
        self.device_info = device_info
        self.territory = TERRITORIES[device_info.used_country_code]
        self.url_prefix = URL_PREFIX[device_info.used_country_code]
        self.device_access.ipaddress = device_info.IPAddress



"""Sky Q app methods and data class."""

import json
from dataclasses import dataclass, field

from ..const import APP_EPG, APP_STATUS_VISIBLE, REST_PATH_APPS, WS_CURRENT_APPS


class AppInformation:
    """Sky Q app information retrieval methods."""

    def __init__(self, remote_config):
        """Initialise the app information class."""
        self._device_access = remote_config.device_access
        self._current_app = APP_EPG
        self._apps = {}

    def get_active_application(self):
        """Get the active application on Sky Q box."""
        try:
            if apps := self._device_access.call_sky_web_socket(WS_CURRENT_APPS):
                self._current_app = next(
                    a for a in apps["apps"] if a["status"] == APP_STATUS_VISIBLE
                )["appId"]

            return App(self._current_app, self._get_app_title(self._current_app))
        except Exception:  # pylint: disable=broad-except
            return App(self._current_app, self._get_app_title(self._current_app))

    def _get_app_title(self, appid):
        if len(self._apps) == 0:
            apps = self._device_access.retrieve_information(REST_PATH_APPS)
            if not apps:
                return None
            for app in apps["apps"]:
                self._apps[app["appId"]] = app["title"]

        return self._apps[appid] if appid in self._apps else None


@dataclass
class App:
    """SkyQ App Class."""

    appId: str = field(  # pylint: disable=invalid-name
        init=True,
        repr=True,
        compare=False,
    )
    title: str = field(
        init=True,
        repr=True,
        compare=False,
    )

    def as_json(self) -> str:
        """Return a JSON string representing this app info."""
        return json.dumps(self, cls=_AppJSONEncoder)


def app_decoder(obj):
    """Decode programme object from json."""
    app = json.loads(obj)
    if "__type__" in app and app["__type__"] == "__app__":
        return App(**app["attributes"])
    return app


class _AppJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, App):
            attributes = dict(vars(o))
            return {
                "__type__": "__app__",
                "attributes": attributes,
            }



"""Structure of a standard EPG programme."""

import json
from collections import OrderedDict
from dataclasses import dataclass, field
from datetime import datetime, timedelta

import requests
from ..const import (
    LIVE_IMAGE_URL,
    RESPONSE_OK,
    SCHEDULE_URL,
    SKY_STATUS_LIVE,
)

from .channel import ChannelInformation, build_channel_image_url
from .programme import Programme


class ChannelEPGInformation:
    """Sky Q Channel EPG information retrieval methods."""

    def __init__(self, remote_config):
        """Initialise the channel epg information class."""
        self._remote_config = remote_config
        self._device_access = remote_config.device_access
        self._remote_country = remote_config.remote_country
        self._test_channel = remote_config.test_channel
        self._epg_cache_len = remote_config.epg_cache_len
        self._epg_cache = OrderedDict()
        self._channel = None
        self._channel_information = None

    def get_epg_data(self, sid, epg_date, days=2):
        """Get EPG data for the specified channel/date."""
        epg = f'{sid} {"{:0>2d}".format(days)} {epg_date.strftime("%Y%m%d")}'

        if sid in self._epg_cache and self._epg_cache[sid]["epg"] == epg:
            return self._epg_cache[sid]["channel"]

        channel_no = None
        channel_name = None
        channel_image_url = None
        programmes = set()

        if channel_node := self._get_channel_node(sid):
            channel_no = channel_node["channelno"]
            channel_name = channel_node["channel"]
            channel_image_url = build_channel_image_url(
                sid,
                channel_name,
                self._remote_config.url_prefix,
                self._remote_config.territory,
            )

            for day in range(days):
                programmes_data = self._get_data(
                    sid, channel_name, epg_date + timedelta(days=day)
                )
                if len(programmes_data) > 0:
                    programmes = programmes.union(programmes_data)
                else:
                    break

        self._channel = ChannelEPG(
            sid, channel_no, channel_name, channel_image_url, sorted(programmes)
        )
        self._epg_cache[sid] = {
            "epg": epg,
            "channel": self._channel,
            "updatetime": datetime.utcnow(),
        }
        self._epg_cache = OrderedDict(
            sorted(
                self._epg_cache.items(), key=lambda x: x[1]["updatetime"], reverse=True
            )
        )
        while len(self._epg_cache) > self._epg_cache_len:
            self._epg_cache.popitem(last=True)

        return self._channel

    def _get_data(self, sid, channel_name, epg_date):
        programmes = set()
        epg_data = self._get_day_epg_data(sid, epg_date)
        if epg_data is None:
            return programmes

        if len(epg_data) == 0:
            return programmes

        for programme in epg_data[0]["events"]:
            starttime = datetime.utcfromtimestamp(programme["st"])
            endtime = datetime.utcfromtimestamp(programme["st"] + programme["d"])
            title = programme["t"]
            season = None
            if "seasonnumber" in programme and programme["seasonnumber"] > 0:
                season = programme["seasonnumber"]
            episode = None
            if "episodenumber" in programme and programme["episodenumber"] > 0:
                episode = programme["episodenumber"]
            programmeuuid = None
            image_url = None
            if "programmeuuid" in programme:
                programmeuuid = str(programme["programmeuuid"])
                image_url = LIVE_IMAGE_URL.format(
                    programmeuuid,
                    self._remote_config.url_prefix,
                    self._remote_config.territory,
                )

            eid = programme["eid"]
            programme = Programme(
                programmeuuid,
                starttime,
                endtime,
                title,
                season,
                episode,
                image_url,
                channel_name,
                SKY_STATUS_LIVE,
                "n/a",
                eid,
            )
            programmes.add(programme)

        return programmes

    def _get_day_epg_data(self, sid, epg_date):
        epg_date_str = epg_date.strftime("%Y%m%d")

        epg_url = SCHEDULE_URL.format(sid, epg_date_str)
        headers = {
            "x-skyott-territory": self._remote_config.territory,
            "x-skyott-provider": "SKY",
            "x-skyott-proposition": "SKYQ",
        }
        resp = requests.get(epg_url, headers=headers)
        return resp.json()["schedule"] if resp.status_code == RESPONSE_OK else None

    def _get_channel_node(self, sid):
        if not self._channel_information:
            self._channel_information = ChannelInformation(self._remote_config)

        return self._channel_information.get_channel_node(sid)


@dataclass
class ChannelEPG:
    """SkyQ Channel EPG Class."""

    sid: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    channelno: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    channelname: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    channelimageurl: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    programmes: set = field(
        init=True,
        repr=True,
        compare=False,
    )

    def as_json(self) -> str:
        """Return a JSON string representing this Channel."""
        return json.dumps(self, cls=_ChannelEPGJSONEncoder)


def channel_epg_decoder(obj):
    """Decode channel object from json."""
    channelepg = json.loads(obj, object_hook=_json_decoder_hook)
    if "__type__" in channelepg and channelepg["__type__"] == "__channelepg__":
        return ChannelEPG(
            programmes=channelepg["programmes"], **channelepg["attributes"]
        )
    return channelepg


def _json_decoder_hook(obj):
    """Decode JSON into appropriate types used in this library."""
    if "starttime" in obj:
        obj["starttime"] = datetime.strptime(obj["starttime"], "%Y-%m-%dT%H:%M:%SZ")
    if "endtime" in obj:
        obj["endtime"] = datetime.strptime(obj["endtime"], "%Y-%m-%dT%H:%M:%SZ")
    if "__type__" in obj and obj["__type__"] == "__programme__":
        obj = Programme(**obj["attributes"])
    return obj


class _ChannelEPGJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, ChannelEPG):
            type_ = "__channelepg__"
            programmes = o.programmes
            attributes = {k: v for k, v in vars(o).items() if k not in {"programmes"}}
            return {
                "__type__": type_,
                "attributes": attributes,
                "programmes": programmes,
            }

        if isinstance(o, set):
            return list(o)

        if isinstance(o, Programme):
            attributes = {}
            for k, val in vars(o).items():
                if isinstance(val, datetime):
                    val = val.strftime("%Y-%m-%dT%H:%M:%SZ")
                attributes[k] = val
            return {
                "__type__": "__programme__",
                "attributes": attributes,
            }

        json.JSONEncoder.default(self, o)  # pragma: no cover



"""Structure of a media information."""

import json
from dataclasses import dataclass, field
from datetime import datetime

from ..const import CURRENT_URI, PVR, UPNP_GET_MEDIA_INFO, XSI
from .channel import ChannelInformation, build_channel_image_url


class MediaInformation:
    """Sky Q media information retrieval methods."""

    def __init__(self, remote_config):
        """Initialise the media information class."""
        self._remote_config = remote_config
        self._device_access = remote_config.device_access
        self._remote_country = remote_config.remote_country
        self._test_channel = remote_config.test_channel
        self._channel_information = None

    def get_current_media(self):
        """Get the currently playing media on the SkyQ box."""
        channel = None
        channelno = None
        image_url = None
        sid = None
        pvrid = None
        live = False

        response = self._device_access.call_sky_soap_service(UPNP_GET_MEDIA_INFO)
        if response is None:
            return None

        current_uri = response[CURRENT_URI]
        if current_uri is None:
            return None

        if XSI in current_uri:
            sid = self._test_channel or int(current_uri[6:], 16)
            live = True
            if channel_node := self._get_channel_node(sid):
                channel = channel_node["channel"]
                channelno = channel_node["channelno"]
                image_url = build_channel_image_url(
                    sid,
                    channel,
                    self._remote_config.url_prefix,
                    self._remote_config.territory,
                )
        elif PVR in current_uri:
            # Recorded content
            pvrid = f"P{current_uri[11:]}"
            live = False

        return Media(channel, channelno, image_url, sid, pvrid, live)

    def _get_channel_node(self, sid):
        if not self._channel_information:
            self._channel_information = ChannelInformation(self._remote_config)

        return self._channel_information.get_channel_node(sid)


@dataclass
class Media:
    """SkyQ Programme Class."""

    channel: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    channelno: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    image_url: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    sid: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    pvrid: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    live: bool = field(
        init=True,
        repr=True,
        compare=False,
    )

    def as_json(self) -> str:
        """Return a JSON string representing this media info."""
        return json.dumps(self, cls=_MediaJSONEncoder)


def media_decoder(obj):
    """Decode programme object from json."""
    media = json.loads(obj)
    if "__type__" in media and media["__type__"] == "__media__":
        return Media(**media["attributes"])
    return media


class _MediaJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Media):
            attributes = {}
            for k, val in vars(o).items():
                if isinstance(val, datetime):
                    val = val.strftime("%Y-%m-%dT%H:%M:%SZ")
                attributes[k] = val
            return {
                "__type__": "__media__",
                "attributes": attributes,
            }



"""SKY Q Remote Utilities."""
import json
import logging
import math
import socket
import time
from http import HTTPStatus

import requests
import websocket
import xmltodict

from ..const import (
    COMMANDS,
    CONNECT_TIMEOUT,
    HTTP_TIMEOUT,
    REST_BASE_URL,
    REST_DELETE,
    REST_GET,
    REST_POST,
    SKY_PLAY_URN,
    SKYCONTROL,
    SOAP_ACTION,
    SOAP_CONTROL_BASE_URL,
    SOAP_DESCRIPTION_BASE_URL,
    SOAP_PAYLOAD,
    SOAP_RESPONSE,
    SOAP_TIMEOUT,
    SOAP_USER_AGENT,
    WS_BASE_URL,
)

_LOGGER = logging.getLogger(__name__)
UNDEFINED = "undefined"


class DeviceAccess:
    """Set up the device for access."""

    def __init__(self, host, json_port, port):
        """Initialise the utility setup."""
        self._host = host
        self._json_port = json_port
        self._port = port
        self.ipaddress = None
        # _LOGGER.debug(f"Init device access - {self._host}")
        self._soap_control_url = UNDEFINED

    def retrieve_information(self, rest_path, call_type=REST_GET):
        """Retrieve information from the SkyQ box."""
        try:
            if call_type == REST_GET:
                return self.http_json(rest_path)
            if call_type == REST_POST:
                return self.http_json_post(rest_path)
            if call_type == REST_DELETE:
                return self.http_json_delete(rest_path)
        except (
            requests.exceptions.ConnectTimeout,
            requests.exceptions.ConnectionError,
            requests.exceptions.ReadTimeout,
        ):  # as err:
            # _LOGGER.debug(f"D0010U - Connection error: {self._host} : {err}")
            return None
        # except Exception as err:
        #     _LOGGER.exception(f"X0010DA - Error occurred: {self._host} : {err}")
        #     return None

    def call_sky_web_socket(self, method):
        """Make a websocket call to the sky box."""
        _LOGGER.debug("WS Call - %s - %s", self._host, method)
        try:
            websock = websocket.create_connection(
                WS_BASE_URL.format(self._host, method)
            )
            response = json.loads(websock.recv())
            websock.close()
            return response
        except (TimeoutError) as err:
            _LOGGER.warning(
                "W0010DA - Websocket call failed: %s : %s : %s", self._host, method, err
            )
            return {"url": None, "status": "Error"}
        except Exception as err:  # pylint: disable=broad-except
            _LOGGER.exception(
                "X0020DA - Error occurred: %s : %s : %s", self._host, method, err
            )
            return None

    def call_sky_soap_service(self, method):
        """Make a SOAP call to the sky box."""
        if self._soap_control_url == UNDEFINED:
            self._soap_control_url = self._get_soap_control_url()
        if not self._soap_control_url:
            _LOGGER.warning(
                "W0020DA - No Control URL, SOAP call not made: %s : %s ",
                self._host,
                method,
            )
            return None
        _LOGGER.debug("SOAP Call - %s : %s", self._host, method)
        try:
            payload = SOAP_PAYLOAD.format(method)
            headers = {
                "Content-Type": 'text/xml; charset="utf-8"',
                "SOAPACTION": SOAP_ACTION.format(method),
            }
            resp = requests.post(
                self._soap_control_url,
                headers=headers,
                data=payload,
                verify=True,
                timeout=SOAP_TIMEOUT,
            )
            if resp.status_code == HTTPStatus.OK:
                xml = resp.text
                return xmltodict.parse(xml)["s:Envelope"]["s:Body"][
                    SOAP_RESPONSE.format(method)
                ]
            return None
        except requests.exceptions.RequestException:
            return None

    def http_json(self, path, headers=None) -> str:
        """Make an HTTP get call to the sky box."""
        _LOGGER.debug("HTTP Get Call - %s - %s", self._host, path)
        response = requests.get(
            REST_BASE_URL.format(self._host, self._json_port, path),
            timeout=HTTP_TIMEOUT,
            headers=headers,
        )
        return response.json()

    def http_json_post(self, path, headers=None) -> str:
        """Make an HTTP post call to the sky box."""
        _LOGGER.debug("HTTP Post Call - %s - %s", self._host, path)
        response = requests.post(
            REST_BASE_URL.format(self._host, self._json_port, path),
            timeout=HTTP_TIMEOUT,
            headers=headers,
        )
        return response.status_code

    def http_json_delete(self, path, headers=None) -> str:
        """Make an HTTP delete call to the sky box."""
        _LOGGER.debug("HTTP Delete Call - %s - %s", self._host, path)
        response = requests.delete(
            REST_BASE_URL.format(self._host, self._json_port, path),
            timeout=HTTP_TIMEOUT,
            headers=headers,
        )
        return response.status_code

    def send_command(self, port, code):
        """Send a command to the sky box."""
        _LOGGER.debug("Socket Call - %s - %s", self._host, code)
        command_bytes = bytearray(
            [4, 1, 0, 0, 0, 0, int(math.floor(224 + (code / 16))), code % 16]
        )

        try:
            client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        except socket.error as err:
            _LOGGER.exception(
                "X0030DA - Failed to create socket when sending command: %s : %s ",
                self._host,
                err,
            )
            return

        try:
            client.connect((self._host, port))
        except Exception as err:  # pylint: disable=broad-except
            _LOGGER.exception(
                "X0040DA - Failed to connect to client when sending command: %s : %s",
                self._host,
                err,
            )
            return

        strlen = 12
        timeout = time.time() + CONNECT_TIMEOUT

        while 1:
            data = client.recv(1024)

            if len(data) < 24:
                client.sendall(data[:strlen])
                strlen = 1
            else:
                client.sendall(command_bytes)
                command_bytes[1] = 0
                client.sendall(command_bytes)
                client.close()
                break

            if time.time() > timeout:
                _LOGGER.error(
                    "E0010DA - Timeout error sending command: %s : %s", self._host, code
                )
                break

    def press(self, sequence):
        """Issue the specified sequence of commands to SkyQ box."""
        if isinstance(sequence, list):
            for item in sequence:
                if item.casefold() not in COMMANDS:
                    _LOGGER.error(
                        "E0020DA - Invalid command: %s : %s", self._host, item
                    )
                    break
                self.send_command(self._port, COMMANDS[item.casefold()])
                time.sleep(0.5)
        elif sequence not in COMMANDS:
            _LOGGER.error("E0030DA - Invalid command: %s : %s", self._host, sequence)
        else:
            self.send_command(self._port, COMMANDS[sequence.casefold()])

    def _find_play_service(self, description):
        services = description["root"]["device"]["serviceList"]["service"]
        if not isinstance(services, list):
            services = [services]
        play_service = None
        for service in services:
            if service["serviceId"] == SKY_PLAY_URN:
                play_service = service

        return play_service

    def _get_soap_control_url(self):
        """Get the soapcontrourl for the SkyQ box."""
        url_index = 0
        soap_control_url = None
        while soap_control_url is None and url_index < 50:
            soap_control_url = self._get_soap_control_url_item(url_index)["url"]
            url_index += 1

        if not soap_control_url:
            _LOGGER.warning("W0030DA - Soap Control URL not available - %s", self._host)

        return soap_control_url

    def _get_soap_control_url_item(self, description_index):
        """Get the sky control url."""
        # _LOGGER.debug("SoapControlURL - %s - %s", self.ipaddress, description_index)
        description_url = SOAP_DESCRIPTION_BASE_URL.format(
            self.ipaddress, description_index
        )
        headers = {"User-Agent": SOAP_USER_AGENT}
        empty_return = {"url": None, "status": "Not Found"}
        try:
            resp = requests.get(description_url, headers=headers, timeout=SOAP_TIMEOUT)
            if resp.status_code == HTTPStatus.OK:
                description = xmltodict.parse(resp.text)
                device_type = description["root"]["device"]["deviceType"]
                if SKYCONTROL not in device_type:
                    return empty_return

                play_service = self._find_play_service(description)

                if play_service is None:
                    return empty_return

                return {
                    "url": SOAP_CONTROL_BASE_URL.format(
                        self._host,
                        play_service[  # pylint: disable=unsubscriptable-object
                            "controlURL"
                        ],
                    ),
                    "status": "OK",
                }
            return empty_return
        except requests.exceptions.Timeout:
            _LOGGER.debug(
                "D0020DA - Control URL not accessible: %s : %s",
                self._host,
                description_url,
            )
            return {"url": None, "status": "Error"}
        except (requests.exceptions.ConnectionError) as err:
            _LOGGER.exception("X0050DA - Connection error: %s : %s", self._host, err)
            return empty_return
        except Exception as err:  # pylint: disable=broad-except
            _LOGGER.exception(
                "X0060DA - Other error occurred: %s : %s", self._host, err
            )
            return empty_return



"""Methods for retrieving recording information."""

import json
import logging
from dataclasses import dataclass, field
from datetime import datetime

from pyskyqremote.classes.channel import build_channel_image_url

from ..const import (ALLRECORDINGS, PVR_IMAGE_URL, RESPONSE_OK,
                     REST_BOOK_PPVRECORDING, REST_BOOK_RECORDING,
                     REST_BOOK_SERIES_RECORDING, REST_DELETE, REST_POST,
                     REST_QUOTA_DETAILS, REST_RECORDING_DELETE,
                     REST_RECORDING_DETAILS, REST_RECORDING_ERASE,
                     REST_RECORDING_ERASE_ALL, REST_RECORDING_KEEP,
                     REST_RECORDING_LOCK,
                     REST_RECORDING_SET_LAST_PLAYED_POSITION,
                     REST_RECORDING_UNDELETE, REST_RECORDING_UNKEEP,
                     REST_RECORDING_UNLOCK, REST_RECORDINGS_LIST,
                     REST_SERIES_LINK, REST_SERIES_UNLINK)
from .programme import Programme

_LOGGER = logging.getLogger(__name__)


class RecordingsInformation:
    """Sky Q recordings information retrieval methods."""

    def __init__(self, remote_config):
        """Initialise the recordings information class."""
        self._remote_config = remote_config

    def get_recordings(self, status, limit, offset):
        """Get the list of available Recordings."""
        recordings = set()
        resp = self._remote_config.device_access.retrieve_information(
            REST_RECORDINGS_LIST.format(limit, offset)
        )
        if not resp or "pvrItems" not in resp:
            _LOGGER.error(
                "E0010R - Timeout retrieving recordings: %s", self._remote_config.host
            )
            return Recordings(recordings)
        rec_data = resp["pvrItems"]
        for recording in rec_data:
            if recording["status"] == status or status == ALLRECORDINGS:
                built = self._build_recording(recording)
                recordings.add(built)

        return Recordings(recordings)

    def get_recording(self, pvrid):
        """Get the recording details."""
        resp = self._remote_config.device_access.retrieve_information(
            REST_RECORDING_DETAILS.format(pvrid)
        )
        if not resp or "details" not in resp:
            _LOGGER.info("I0010R - Recording data not found for %s", pvrid)
            return None

        recording = resp["details"]

        return self._build_recording(recording)

    def get_quota(self):
        """Get the quota information."""
        resp = self._remote_config.device_access.retrieve_information(
            REST_QUOTA_DETAILS
        )
        if not resp:
            return None

        if "userQuotaMax" not in resp:
            _LOGGER.debug("D0010R - Recording data not found for %s", resp)
            return None
        return Quota(resp["userQuotaMax"], resp["userQuotaUsed"])

    def book_recording(self, eid, series):
        """Book recording for specified item."""
        resp = None
        if series:
            resp = self._remote_config.device_access.retrieve_information(
                REST_BOOK_SERIES_RECORDING.format(eid), REST_POST
            )
        else:
            resp = self._remote_config.device_access.retrieve_information(
                REST_BOOK_RECORDING.format(eid), REST_POST
            )

        if resp != RESPONSE_OK:
            return False

        return True

    def book_ppv_recording(self, eid, offerref):
        """Book PPV recording for specified item."""
        resp = self._remote_config.device_access.retrieve_information(
            REST_BOOK_PPVRECORDING.format(eid, offerref), REST_POST
        )
        if resp != RESPONSE_OK:
            return False

        return True

    def series_link(self, pvrid, linkon):
        """Series link the specified item."""
        resp = None
        if linkon:
            resp = self._remote_config.device_access.retrieve_information(
                REST_SERIES_LINK.format(pvrid), REST_POST
            )
        else:
            resp = self._remote_config.device_access.retrieve_information(
                REST_SERIES_UNLINK.format(pvrid), REST_POST
            )

        if resp != RESPONSE_OK:
            return False

        return True

    def recording_keep(self, pvrid, keepon):
        """Keep the specified item."""
        resp = None
        if keepon:
            resp = self._remote_config.device_access.retrieve_information(
                REST_RECORDING_KEEP.format(pvrid), REST_POST
            )
        else:
            resp = self._remote_config.device_access.retrieve_information(
                REST_RECORDING_UNKEEP.format(pvrid), REST_POST
            )

        if resp != RESPONSE_OK:
            return False

        return True

    def recording_lock(self, pvrid, lockon):
        """Lock the specified item."""
        resp = None
        if lockon:
            resp = self._remote_config.device_access.retrieve_information(
                REST_RECORDING_LOCK.format(pvrid), REST_POST
            )
        else:
            resp = self._remote_config.device_access.retrieve_information(
                REST_RECORDING_UNLOCK.format(pvrid), REST_POST
            )

        if resp != RESPONSE_OK:
            return False

        return True

    def recording_delete(self, pvrid, deleteon):
        """Delete the specified item."""
        resp = None
        if deleteon:
            resp = self._remote_config.device_access.retrieve_information(
                REST_RECORDING_DELETE.format(pvrid), REST_POST
            )
        else:
            resp = self._remote_config.device_access.retrieve_information(
                REST_RECORDING_UNDELETE.format(pvrid), REST_POST
            )

        if resp != RESPONSE_OK:
            return False

        return True

    def recording_erase(self, pvrid):
        """Permanently erase the specified item."""
        resp = self._remote_config.device_access.retrieve_information(
            REST_RECORDING_ERASE.format(pvrid), REST_POST
        )

        if resp != RESPONSE_OK:
            return False

        return True

    def recording_erase_all(self):
        """Permanently erase the specified item."""
        resp = self._remote_config.device_access.retrieve_information(
            REST_RECORDING_ERASE_ALL, REST_DELETE
        )

        if resp != RESPONSE_OK:
            return False

        return True

    def recording_set_last_played_position(self, pvrid, pos):
        """Set the last played position for specified item."""
        resp = self._remote_config.device_access.retrieve_information(
            REST_RECORDING_SET_LAST_PLAYED_POSITION.format(pos, pvrid), REST_POST
        )
        if resp != RESPONSE_OK:
            return False

        return True

    def _build_recording(self, recording):
        channel = recording["cn"]
        title = recording["t"]

        season = None
        episode = None
        if "seasonnumber" in recording and "episodenumber" in recording:
            season = recording["seasonnumber"]
            episode = recording["episodenumber"]

        programmeuuid = None
        image_url = None
        if "programmeuuid" in recording:
            programmeuuid = recording["programmeuuid"]
            image_url = PVR_IMAGE_URL.format(
                programmeuuid,
                self._remote_config.url_prefix,
                self._remote_config.territory,
            )
        elif "osid" in recording:
            sid = str(recording["osid"])
            image_url = build_channel_image_url(
                sid,
                channel,
                self._remote_config.url_prefix,
                self._remote_config.territory,
            )

        starttimestamp = 0
        if "ast" in recording:
            starttimestamp = recording["ast"]
        elif "st" in recording:
            starttimestamp = recording["st"]
        starttime = datetime.utcfromtimestamp(starttimestamp)

        endtime = None
        if "finald" in recording:
            endtime = datetime.utcfromtimestamp(starttimestamp + recording["finald"])
        elif "schd" in recording:
            endtime = datetime.utcfromtimestamp(starttimestamp + recording["schd"])
        else:
            endtime = starttime

        pvrid = recording["pvrid"]

        eid = recording["oeid"] if "oeid" in recording else None
        status = recording["status"]

        return Programme(
            programmeuuid,
            starttime,
            endtime,
            title,
            season,
            episode,
            image_url,
            channel,
            status,
            pvrid,
            eid,
        )


@dataclass
class Recordings:
    """SkyQ Channel EPG Class."""

    programmes: set = field(
        init=True,
        repr=True,
        compare=False,
    )

    def as_json(self) -> str:
        """Return a JSON string representing this Channel."""
        return json.dumps(self, cls=_RecordingsJSONEncoder)


def recordings_decoder(obj):
    """Decode channel object from json."""
    recordings = json.loads(obj, object_hook=_json_decoder_hook)
    if "__type__" in recordings and recordings["__type__"] == "__recordings__":
        return Recordings(
            programmes=recordings["programmes"], **recordings["attributes"]
        )
    return recordings


def _json_decoder_hook(obj):
    """Decode JSON into appropriate types used in this library."""
    if "starttime" in obj:
        obj["starttime"] = datetime.strptime(obj["starttime"], "%Y-%m-%dT%H:%M:%SZ")
    if "endtime" in obj:
        obj["endtime"] = datetime.strptime(obj["endtime"], "%Y-%m-%dT%H:%M:%SZ")
    if "__type__" in obj and obj["__type__"] == "__programme__":
        obj = Programme(**obj["attributes"])
    return obj


class _RecordingsJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Recordings):
            type_ = "__recordings__"
            programmes = o.programmes
            attributes = {k: v for k, v in vars(o).items() if k not in {"programmes"}}
            return {
                "__type__": type_,
                "attributes": attributes,
                "programmes": programmes,
            }

        if isinstance(o, set):
            return list(o)

        if isinstance(o, Programme):
            attributes = {}
            for k, val in vars(o).items():
                if isinstance(val, datetime):
                    val = val.strftime("%Y-%m-%dT%H:%M:%SZ")
                attributes[k] = val
            return {
                "__type__": "__programme__",
                "attributes": attributes,
            }

        json.JSONEncoder.default(self, o)  # pragma: no cover


@dataclass
class Quota:
    """SkyQ Quota Class."""

    quota_max: int = field(
        init=True,
        repr=True,
        compare=False,
    )
    quota_used: str = field(
        init=True,
        repr=True,
        compare=False,
    )

    def as_json(self) -> str:
        """Return a JSON string representing this quota info."""
        return json.dumps(self, cls=_QuotaJSONEncoder)


def quota_decoder(obj):
    """Decode quota object from json."""
    quota = json.loads(obj)
    if "__type__" in quota and quota["__type__"] == "__quota__":
        return Quota(**quota["attributes"])
    return quota


class _QuotaJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Quota):
            attributes = dict(vars(o))
            return {
                "__type__": "__quota__",
                "attributes": attributes,
            }



"""Methods for retrieving device information."""

import json
import logging
from dataclasses import dataclass, field

from ..const import (CURRENT_SPEED, CURRENT_TRANSPORT_STATE,
                     CURRENT_TRANSPORT_STATUS, DEFAULT_TRANSPORT_SPEED,
                     DEFAULT_TRANSPORT_STATE, KNOWN_COUNTRIES,
                     REST_PATH_DEVICEINFO, REST_PATH_SYSTEMINFO,
                     SKY_STATE_NOMEDIA, SKY_STATE_OFF, SKY_STATE_PAUSED,
                     SKY_STATE_PLAYING, SKY_STATE_STANDBY, SKY_STATE_STOPPED,
                     SKY_STATE_TRANSITIONING, SKY_STATE_UNSUPPORTED,
                     UPNP_GET_TRANSPORT_INFO)

_LOGGER = logging.getLogger(__name__)


class DeviceInformation:
    """Sky Q device information retrieval methods."""

    def __init__(self, remote_config):
        """Initialise the device information class."""
        self._remote_config = remote_config
        self._device_access = remote_config.device_access
        self._port = remote_config.port

    def get_transport_information(self):
        """Get the transport information from the SkyQ box."""
        response = self._device_access.call_sky_soap_service(UPNP_GET_TRANSPORT_INFO)
        if response is not None:
            return TransportInfo(
                response[CURRENT_TRANSPORT_STATE],
                response[CURRENT_TRANSPORT_STATUS],
                response[CURRENT_SPEED],
            )

        return TransportInfo(
            SKY_STATE_OFF, DEFAULT_TRANSPORT_STATE, DEFAULT_TRANSPORT_SPEED
        )

    def get_system_information(self):
        """Get the system information from the SkyQ box."""
        return self._device_access.retrieve_information(REST_PATH_SYSTEMINFO)

    def get_device_information(self, override_country):
        """Get the device information from the SkyQ box."""
        device_info = self._device_access.retrieve_information(REST_PATH_DEVICEINFO)
        if not device_info:
            return None

        system_info = self.get_system_information()
        as_version = device_info["ASVersion"]
        ip_address = device_info["IPAddress"]
        country_code = device_info["countryCode"]
        gateway = device_info["gateway"]
        hardware_model = system_info["hardwareModel"]
        device_type = system_info["deviceType"]
        hardware_name = device_info["hardwareName"]
        manufacturer = system_info["manufacturer"]
        model_number = device_info["modelNumber"]
        serial_number = device_info["serialNumber"]
        version_number = device_info["versionNumber"]
        bouquet = device_info["bouquet"]
        subbouquet = device_info["subbouquet"]
        wake_reason = system_info["wakeReason"]
        system_uptime = system_info["systemUptime"]
        hdr_capable = system_info["hdrCapable"]
        uhd_capable = system_info["uhdCapable"]

        used_country_code = override_country or country_code.upper()
        if not used_country_code:
            _LOGGER.error("E0010 - No country identified: %s", self._remote_config.host)
            return None

        if used_country_code in KNOWN_COUNTRIES:
            used_country_code = KNOWN_COUNTRIES[used_country_code]
        else:
            _LOGGER.error(
                "W0010 - Country code %s unknown, defaulting to GBR: %s",
                used_country_code,
                self._remote_config.host,
            )
            used_country_code = "GBR"

        return Device(
            as_version,
            ip_address,
            country_code,
            used_country_code,
            hardware_model,
            hardware_name,
            device_type,
            gateway,
            manufacturer,
            model_number,
            serial_number,
            version_number,
            bouquet,
            subbouquet,
            wake_reason,
            system_uptime,
            hdr_capable,
            uhd_capable,
        )


@dataclass
class Device:
    """SkyQ Device Class."""

    ASVersion: str = field(  # pylint: disable=invalid-name
        init=True,
        repr=True,
        compare=False,
    )
    IPAddress: str = field(  # pylint: disable=invalid-name
        init=True,
        repr=True,
        compare=False,
    )
    countryCode: str = field(  # pylint: disable=invalid-name
        init=True,
        repr=True,
        compare=False,
    )
    used_country_code: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    hardwareModel: str = field(  # pylint: disable=invalid-name
        init=True,
        repr=True,
        compare=False,
    )
    hardwareName: str = field(  # pylint: disable=invalid-name
        init=True,
        repr=True,
        compare=False,
    )
    deviceType: str = field(  # pylint: disable=invalid-name
        init=True,
        repr=True,
        compare=False,
    )
    gateway: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    manufacturer: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    modelNumber: str = field(  # pylint: disable=invalid-name
        init=True,
        repr=True,
        compare=False,
    )
    serialNumber: str = field(  # pylint: disable=invalid-name
        init=True,
        repr=True,
        compare=False,
    )
    versionNumber: str = field(  # pylint: disable=invalid-name
        init=True,
        repr=True,
        compare=False,
    )
    bouquet: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    subbouquet: str = field(
        init=True,
        repr=True,
        compare=False,
    )
    wakeReason: str = field(  # pylint: disable=invalid-name
        init=True,
        repr=True,
        compare=False,
    )
    systemUptime: str = field(  # pylint: disable=invalid-name
        init=True,
        repr=True,
        compare=False,
    )
    hdrCapable: str = field(  # pylint: disable=invalid-name
        init=True,
        repr=True,
        compare=False,
    )
    uhdCapable: str = field(  # pylint: disable=invalid-name
        init=True,
        repr=True,
        compare=False,
    )

    def as_json(self) -> str:
        """Return a JSON string representing this device info."""
        return json.dumps(self, cls=_DeviceJSONEncoder)


def device_decoder(obj):
    """Decode programme object from json."""
    device = json.loads(obj)
    if "__type__" in device and device["__type__"] == "__device__":
        return Device(**device["attributes"])
    return device


class _DeviceJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Device):
            attributes = dict(vars(o))
            return {
                "__type__": "__device__",
                "attributes": attributes,
            }


@dataclass
class TransportInfo:
    """SkyQ TransportInfo Class."""

    CurrentTransportState: str = field(  # pylint: disable=invalid-name
        init=True,
        repr=True,
        compare=False,
    )
    CurrentTransportStatus: str = field(  # pylint: disable=invalid-name
        init=True,
        repr=True,
        compare=False,
    )
    CurrentSpeed: float = field(  # pylint: disable=invalid-name
        init=True,
        repr=True,
        compare=False,
    )
    state: str = None

    def __post_init__(self):
        """Post process the transport setup."""
        self.state = self.CurrentTransportState
        if self.CurrentTransportState in [SKY_STATE_NOMEDIA, SKY_STATE_STOPPED]:
            self.state = SKY_STATE_STANDBY
        if self.CurrentTransportState in [SKY_STATE_PLAYING, SKY_STATE_TRANSITIONING]:
            self.state = SKY_STATE_PLAYING
        if self.CurrentTransportState == SKY_STATE_PAUSED:
            self.state = SKY_STATE_PAUSED
        if self.CurrentTransportState == SKY_STATE_UNSUPPORTED:
            self.CurrentTransportState = None

    def as_json(self) -> str:
        """Return a JSON string representing this app info."""
        return json.dumps(self, cls=_TransportInfoJSONEncoder)


def transportinfo_decoder(obj):
    """Decode programme object from json."""
    transportinfo = json.loads(obj)
    if "__type__" in transportinfo and transportinfo["__type__"] == "__transportinfo__":
        return TransportInfo(**transportinfo["attributes"])
    return transportinfo


class _TransportInfoJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, TransportInfo):
            attributes = dict(vars(o))
            return {
                "__type__": "__transportinfo__",
                "attributes": attributes,
            }
